/*
  -------------------- file back1int.c
  Created 21 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.

  Interface between application, and internal "back1" search routines.
  Internal back1 functions are here, since the whole thing is small.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "our_common.h"
#include "../libinclude/ntup.h"
#include "../libinclude/tuputil.h"
#include "../general/coord.h"
#include "../general/backgen.h"
#include "../general/run.h"
#include "../general/pi.h"
#include "../general/teval.h"
#include "../prog/memalloc.h"
#include "../prog/running.h"
#include "../prog/vector.h"

#define UTEST_INTERVAL  99  /* check for update every UTEST_INTERVAL nodes */

/* #define RANDOM_FIXED */

/*  ---------- Data used by back1_work() and its descendants.
    These things are really parameters to back1_work(), but since
    we do not have control here over how back1_work is called, static
    variables are used instead.

    Initialized in ntup_find_back1.
*/

static FILE *s_f;
static int s_update_interval;
static unsigned s_ioflags;
static double s_max_nodes;
static int s_action;
static int s_print;
static int s_n;
static int s_tuple_kind;
static double *s_lev_nodes;   /* number of nodes processed at each level */

static int
getcolor(
const int cindex,
const int id,
const int * const id_to_first_color
)
{
  if (cindex == 0)
    return(id_to_first_color[id]);
  else if (cindex == 1)
    return(1-id_to_first_color[id]);
  else
    fatal("getcolor()\n");
  DO_GCCHAPPY(return(-1);)
}

/*  Convert between n-tuple formats.  Source format is the one used by
    the back1 backtracker.  Destination format is an array of pix_struct
    structures.
*/
static void
convert_to_pixlist(
const int len,                /* information about "len" pixels */
const int *ids,               /* ids of tuple pixels */
const int *color_indices,     /* color indices of tuple pixels */
const int *id_to_row,         /* map from ids to rows */
const int *id_to_col,         /* map from ids to columns */
const int *id_to_first_color, /* map from ids to starting colors */
struct pix_struct *p_ret      /* assumed to have space for >= len of these */
)
{
  int i;
  for (i = 0; i < len; i++) {
    const int id = ids[i];
    struct pix_struct *const p = & p_ret[i];
    p->row = id_to_row[id];
    p->col = id_to_col[id];
    p->color = getcolor(color_indices[i],id,id_to_first_color);
  }
}

/*  Determine whether a given n-tuple is canonical.
    Returns 0 if not canonical
            1 if canonical
*/
static int
is_tuple_canonical(
struct pix_struct *pixels,
const int len
)
{
  int i;
  int foundrow, foundcol;

  foundrow = foundcol = 0;
  for (i = 0; i < len; i++) {
    const struct pix_struct *p = & pixels[i];
    if (p->row == 0)  foundrow = 1;
    if (p->col == 0)  foundcol = 1;
  }
  if (foundrow && foundcol)  return(1);
  return(0);
}

/*  Sets *status_ret to WORKFUNC_XX
    If we found a solution, we put the solution in *sol_ret.

    Assumes that the caller has set up *sol_ret to hold a tuple
    of size n.
*/
static void
back1_work(
const int original_p,
const int q,
const int do_maxp,              /* should we find max p for given q? */
const struct tuple_struct * const starting_tuple,   /* NULL indicates none */
                                                    /* we ignore */
const struct coord_struct * const pi,
const struct tmask_struct * const pos_tmask,
const struct tmask_struct * const neg_tmask,
struct tuple_struct * const solution_ret,
int * const status_ret
)
{
  int p = original_p;
  const int verbose = (s_action == BACK_ACTION_COUNT) ||
                                      (s_update_interval != UPDATE_NONE);
  struct coord_struct randpi[1];
  int *id_to_first_color;         /* indexed by randid */

  /* current solution */
  int *randids;
  int *color_indices;

  int max_id;         /* max square id (number of ids in pi - 1) */
  int max_cindex;     /* max color index (0 or 1) */
  int k;              /* search level (0 .. n-1) */

  double stats_nfound;          /* number of solutions found so far */
  double stats_nfound_canon;    /* number of canonical solutions found so far */
  double stats_search_nodes;    /* number of search nodes so far */

  int update_counter;   /* for not wasting cpu time on calls to update_test */
  struct pix_struct *pixbuff;

  bsbegin();

  teval_setup_pq(p,q);
  teval_setup_invariants(s_n, pos_tmask, neg_tmask, bsalloc);

  s_lev_nodes = vec_d_alloc(s_n, bsalloc);
  vec_d_zero(s_lev_nodes, s_n);
  pixbuff = bsalloc(s_n * sizeof(struct pix_struct));

  max_id = pi->nids-1;
  id_to_first_color = vec_alloc(max_id+1, bsalloc);
  randids = vec_alloc(s_n, bsalloc);
  color_indices = vec_alloc(s_n, bsalloc);

  if (verbose) {
    fprintf(s_f,"Starting back1 search.\n");
    fprintf(s_f,"  n=%d  p=%d  q=%d  action=%s  max nodes=%g  printing ",
      s_n,p,q, s_action==BACK_ACTION_ONE ? "one" : "count",
      s_max_nodes);
    switch (s_print) {
      case (BACK_PRINT_NONE):       fprintf(s_f, "none.\n"); break;
      case (BACK_PRINT_FIRST):      fprintf(s_f, "first.\n"); break;
      case (BACK_PRINT_ALL):        fprintf(s_f, "all.\n"); break;
      case (BACK_PRINT_CANON_ONLY): fprintf(s_f, "canonical.\n"); break;
    }
    if (starting_tuple)
      fprintf(s_f, "Starting tuple specified; ignoring it.\n");
    if (s_update_interval != UPDATE_NONE) {
      update_init();
      update_set_interval(s_update_interval);
      update_start_timer();
    }
  }

#ifdef RANDOM_FIXED
  {
    int srandom P(( int ));
    srandom(p*q*379+2013);
    printf("WARNING: using RANDOM_FIXED.\n");
  }
#endif  /*RANDOM_FIXED*/

  /* generate a random numbering of pi */
  pi_make_random(randpi, pi, bsalloc);

  DO_GCCHAPPY(max_cindex = 0;)
  switch (s_tuple_kind) {
    case (TUPLE_BLACKONLY):
      max_cindex = 0;
      vec_set(id_to_first_color,max_id+1,BLACK);
      break;
    case (TUPLE_WHITEONLY):
      max_cindex = 0;
      vec_set(id_to_first_color,max_id+1,WHITE);
      break;
    case (TUPLE_BLACKWHITE): {
      long random();
      int i;
      for (i = 0; i <= max_id; i++)  id_to_first_color[i] = random() % 2;
      max_cindex = 1;
      break;
    }
  }
#if defined(DEBUG) || defined(RANDOM_FIXED)
  printf("randpi:\n");
  debug_coord(randpi);
  printf("max_cindex=%d\n", max_cindex);
  printf("id_to_first_color: ");
  vec_print(stdout,id_to_first_color,max_id+1);
#endif  /*DEBUG || RANDOM_FIXED*/


  ASSERT(s_n>0);
  k = 0;                                    /* level 0 */
  randids[0] = 0;                           /* start at pixel id 0 */
  color_indices[0] = 0;                     /* start at color id 0 */
  stats_nfound = stats_nfound_canon = 0;    /* no solutions found yet */
  stats_search_nodes = 0;
  update_counter = 0;

  /*  Note: whenever we reach the top of the loop, we are at a new
      search node in the backtrack search tree.
  */
  while (1) {
    int passed;     /* does current (k+1)-tuple satisfies p,q constraints? */

    ASSERT((k >= 0) && (k < s_n));
    stats_search_nodes++;
    s_lev_nodes[k]++;

#ifdef DEBUG
    printf("----------------------------- top of loop\n");
    printf("ids:           ");
    vec_print(stdout,randids,k+1);
    printf("color_indices: ");
    vec_print(stdout,color_indices,k+1);
#endif  /*DEBUG*/


    /* Periodic status update.  update_test() is inefficient, so we
       take pains to only call it once in a while through the loop.
    */
    if (s_update_interval != UPDATE_NONE) {
      int do_update = 0;
      if (s_update_interval == UPDATE_OFTEN)
        do_update = 1;
      else if (s_update_interval > 0) {
        if (++update_counter >= UTEST_INTERVAL)
          {  update_counter = 0;  do_update = update_test();  }
      }
      if (do_update) {
        int j;
        if (s_ioflags & IOF_NO_TIME_DEPENDENT_OUTPUT)
          fprintf(s_f,"----- Update\n");
        else
          update_header_short(s_f);
        fprintf(s_f,"Working on p=%d, q=%d.",p,q);
        fprintf(s_f,"  At search node %g.  Tuple:\n", stats_search_nodes);
        {
          bsbegin();
          {
            struct pix_struct * const pixels =
                                    bsalloc(s_n*sizeof(struct pix_struct));
            convert_to_pixlist(k+1,randids,color_indices,
              randpi->id_to_row,randpi->id_to_col,id_to_first_color,pixels);
            ntup_print(s_f,pixels,k+1,1); /* 1=long format */
            bsend();
          }
        }
        fprintf(s_f,"ids:          ");
        for (j = 0; j < k+1; j++)
          fprintf(s_f,"%4d", randids[j]);
        fprintf(s_f,"\n");
        fprintf(s_f,"color indices:");
        for (j = 0; j < k+1; j++)
          fprintf(s_f,"%4d", color_indices[j]);
        fprintf(s_f,"\n");
        update_start_timer();
      }
    }

    /* See if current partial solution satisfies p and q constraints
    */

    {
      const int randid  = randids[k];                 /* id in random pi */
      const int color   = getcolor(color_indices[k],randid,id_to_first_color);
      const int row     = randpi->id_to_row[randid];
      const int col     = randpi->id_to_col[randid];
      const int id      = pi->rowcol_to_id[row][col]; /* id in standard pi */
      const int piidc   = PIID_C_TO_PIIDC(id,color);
      passed = do_tupdate(k, piidc);
    }

#ifdef OLD
    in another place above: convert_to_pixlist(k+1,randids,color_indices,
      randpi->id_to_row,randpi->id_to_col,id_to_first_color,pixels);
    if (tmask_pos_satisfied(pixels,k+1,p,pos_tmask)) {
      D(printf("positive test passed.\n");)
      if (tmask_neg_satisfied(pixels,k+1,q,neg_tmask)) {
        D(printf("negative test passed.\n");)
        passed = 1;
      }
      else {
        D(printf("negative test failed.\n");)
      }
    }
    else {
      D(printf("positive test failed.\n");)
    }
#endif /*OLD*/

    D(printf("passed=%d\n", passed);)

    /* process full solution, if found
    */
    if (passed && (k == s_n-1)) {   /* found solution */ 
      int is_canonical;

      convert_to_pixlist(s_n,randids,color_indices,
        randpi->id_to_row,randpi->id_to_col,id_to_first_color,pixbuff);
      is_canonical = is_tuple_canonical(pixbuff,k+1);

      stats_nfound++;
      stats_nfound_canon += is_canonical;

      switch (s_print) {
        case BACK_PRINT_NONE:
          break;
        case BACK_PRINT_FIRST:
          if (stats_nfound == 1) {
            fprintf(s_f, "First solution:\n");
            ntup_print(s_f,pixbuff,k+1,1); /* 1=long format */
          }
          break;
        case BACK_PRINT_ALL:      
          fprintf(s_f, "Solution:\n");
          ntup_print(s_f,pixbuff,k+1,1); /* 1=long format */
          break;
        case BACK_PRINT_CANON_ONLY:
          if (is_canonical) {
            fprintf(s_f, "Canonical solution:\n");
            ntup_print(s_f,pixbuff,k+1,1); /* 1=long format */
          }
          break;
      }

      if (stats_nfound == 1) {      /* store first solution we find */
        int i;
        ASSERT(solution_ret->n == s_n);
        for (i = 0; i < s_n; i++) {
          pix_copy(& solution_ret->pixels[i],& pixbuff[i]);
        }
        solution_ret->p = p;
        solution_ret->q = q;
        if (do_maxp)  {
          *status_ret = WORKFUNC_PMAX;
          for (p++; tmask_pos_satisfied(solution_ret->pixels,
                                      solution_ret->n,p,pos_tmask); p++)  {}
        }
        else {
          *status_ret = WORKFUNC_P;  break;
        }
#ifdef SEE_IF_THIS_IS_NEEDED
        if (s_action == BACK_ACTION_ONE)    /* quit after one solution */
          break;
#endif /*SEE_IF_THIS_IS_NEEDED*/
      }
    }

    /* generate next partial solution, if any
    */
    if ((s_max_nodes > 0) && (stats_search_nodes >= s_max_nodes)) {
      if (verbose)
        fprintf(s_f, "Maximum number of search nodes reached.\n");
      if (stats_nfound == 0)
        *status_ret = WORKFUNC_NOTFOUND;
      break;
    }
    {
      int more;   /* whether or not there are any more partial solutions */
      backgen_next(max_id,max_cindex,s_n,passed,&k,randids,color_indices,&more);
      if (! more) {
        if (verbose)
          fprintf(s_f, "Search space exhausted.\n");
        if (stats_nfound == 0) {
          *status_ret = WORKFUNC_NONEEXIST;
        }
        break;
      }
    }
  }           /* end of main loop */


  if (verbose)
    fprintf(s_f, "Back1 search completed.  %g nodes.\n",
                                                    stats_search_nodes);
  if (s_action == BACK_ACTION_COUNT) {
    fprintf(s_f, "%g solutions, %g canonical solutions.\n",
      stats_nfound, stats_nfound_canon);
  }

  if (s_ioflags & IOF_REPORT_RESULTS)
    lev_nodes_print(s_f,s_lev_nodes,s_n,p,q);

  bsend();
}

/*
  Interface between user and back1 search functions.
  This is what the user calls.
  It is declared in ../libinclude/ntup.h, and not back1int.h as it would be
  if we following the usual convention of placing the function declaration
  in the .h file corresponding to the .c file in which the function resides.

  This function should free all memory it allocates if it sets
  the status field to anything other than NTUP_OK.
*/
void
ntup_find_back1(
const struct to_ntup_struct * const to,
struct from_ntup_struct * const from,
const int     back_action,          /* BACK_ACTION_XX */
const int     back_print,           /* BACK_PRINT_XX */
/* maximum number of nodes before stopping */
const double  back_max_nodes        /* must be 0 or more; 0 means infinity */
)
{
  D(printf("---------------- entering ntup_find_back1\n");)

  /*  Initialize bit manipulation package.  bit_global_init() does
      virtually nothing if it has been called previously.
  */
  bit_global_init();

  /*  -------------------------- Check input for errors, report
                                 input parameters if reporting is specified.
   */
  ntup_check_input(to, & from->status);
  if (from->status == NTUP_OK) {
    ntup_check_input_back(back_action,back_print,back_max_nodes,& from->status);
  }
  if (from->status != NTUP_OK)  return;

  /* Store data in static variables as needed.
  */
  s_f               = to->f;
  s_update_interval = to->update_interval;
  s_ioflags         = to->ioflags;
  s_action          = back_action;
  s_print           = back_print;
  s_max_nodes       = back_max_nodes;
  s_n               = to->n;
  s_tuple_kind      = to->tuple_kind;

  if (s_ioflags & IOF_REPORT_INPUT) {
    fprintf(s_f,"---------- Begin report of back1 input.\n");
    if (back_max_nodes == 0)
      fprintf(s_f, "max_nodes: infinity\n");
    else
      fprintf(s_f, "max_nodes: %g\n", back_max_nodes);
    report_back(s_f,back_action,back_print);
    report_input(to);
    fprintf(s_f,"---------- End report of back1 input.\n\n");
  }

  /* --------------------------------
      Call back1_work() with p and q as needed, keep track of solutions.
      Report generator output if specified.
  */
  ntup_run(to,0,from,back1_work);

  D(printf("---------------- exiting ntup_find_back1\n");)
}

