/*
  -------------------- file teval.c
  Created 13 Jul 94 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* #define DEBUG */
#include "our_common.h"
#include "teval.h"
#include "../libinclude/ntup.h"
#include "../general/coord.h"
#include "../general/pi.h"
#include "prog/bitarray.h"
#include "prog/bitpar.h"
#include "prog/bitword.h"
#include "prog/matrix.h"
#include "prog/memalloc.h"
#include "prog/sort.h"
#include "prog/vector.h"

#include "float.h"

#define QVEC_ROWS   2   /* may be hardcoded in places */
/*COULDDO: investigate whether this can be made 1 without performance loss.
  Guess is that at most a 20% improvement would result.  Might make
  evaluation function worse and thus cause performance loss.
*/

static int s_n;
static int s_p;
static int s_q;
static int s_bitpar_levs;
static const struct tmask_struct * s_pos_tmask;
static const struct tmask_struct * s_neg_tmask;
static int s_nposchars;
static int s_nnegchars;
static const int *s_pos_ntrans_words;
static const int *s_neg_ntrans_words;
static int s_lastlev;       /* level of latest evaluation */

/* COULDDO: move into a single structure for each level.
*/
static BITWORD ***s_pos_mask;
  /* [lev: -1..s_n-1][ch][translation word] */
static BITWORD ****s_neg_count;  /* [lev][ch][translation word][bitpar word] */
static int *s_pvals;      /* pval vector of latest evaluation */
static unsigned *s_ps;    /* piidc's of partial solution that was evaluated */

#if defined(DEBUG_EVAL) || defined(DEBUG)
void
debug_teval(
const char * const str,               /* can be NULL */
const struct teval_struct * const  e
)
{
  if (str)
    printf("-- evaluation for '%s'\n", str);
  printf("ps: ");
  vec_u_print(stdout,e->ps, e->pslen);
  printf("tiebreak=%u\n", e->tiebreak);
  printf("pvals: ");
  vec_print(stdout, e->pvals, s_nposchars);
  printf("qvecs:\n");
  mat_print2(stdout, e->qvecs, QVEC_ROWS, s_nnegchars);
  printf("\n");
}
#endif  /* DEBUG_EVAL || DEBUG */

/*  Give an upper bound on the number of candidates.
    This is the number of piidc's in the entire pi region.
*/
int
can_determine_max(
const struct coord_struct * const pi
)
{
  return(pi->nids * COLORS);
}

/*  Allocate space in the fields of a teval_struct.
    You must have called teval_setup_invariants() before this.
*/
void
teval_alloc(
struct teval_struct * const e,
void  *(*allocfunc) P((const int))
)
{
  e->ps    = vec_u_alloc(s_n, allocfunc);
  e->pvals = vec_alloc(s_nposchars, allocfunc);
  e->qvecs = mat_alloc2(QVEC_ROWS, s_nnegchars, allocfunc);
}

/*  Copy the data pointed to by the fields of a teval struct.
    Do not change the pointers.
*/
void
teval_copy(
struct teval_struct * const dest,
const struct teval_struct *src
)
{
  vec_copy(dest->pvals, src->pvals, s_nposchars);
  mat_copy2(src->qvecs, QVEC_ROWS, s_nnegchars, dest->qvecs);
  dest->pslen     = src->pslen;
  dest->tiebreak  = src->tiebreak;
  vec_u_copy(dest->ps, src->ps, dest->pslen);
}

/*  Given the current partial solution T, for each positive character c there
    is an associated value P(c) that gives the number of positions in which
    c matches T exacty.

    Returns 0 if there exists a character c such that P(c) < p.

    Otherwise, returns 1, and computes P(c) for each positive character.
*/
static int
pos_update(
BITWORD * const * const masks,  /* mask[piid][color], indexed as [ch][word] */
BITWORD * const * const before, /* [ch][word] */
BITWORD * const * const after   /* [ch][word] */
)
{
  int ch, word, count;

  for (ch = 0; ch < s_nposchars; ch++) {
    const int nwords = s_pos_ntrans_words[ch];
    const BITWORD * const our_masks = masks[ch];
    const BITWORD * const our_before = before[ch];
    BITWORD * const our_after = after[ch];
    for (count = 0, word = 0; word < nwords; word++) {
      if (IS_BITWORD_ALL_ZERO(our_before[word])) {
        our_after[word] = ALL_BITS_ZERO;
        /* leave count unchanged */
      }
      else {
        const BITWORD aft = our_masks[word] & our_before[word];
        our_after[word] = aft;
        count += bitword_count(aft);
      }
    }
    if (count < s_p)
      return(0);                /* failed */
    s_pvals[ch] = count;
  }

  return(1);                    /* passed */
}

/*  Returns 0 if failed, 1 if success.
    If successful we set after[0 .. nwords-1][].
    If unsuccessul then after[] is undefined.
*/
static int
neg_update_inner(
const int nwords,               /* num. of translation words */
const BITWORD * const ourmask,  /* mask[piid][color][ch], indexed as [word] */
BITWORD * const * const before, /* [tword][bitpar word] */
BITWORD * const * const after   /* [tword][bitpar word] */
)
{
  int word;

  for (word = 0; word < nwords; word++) {
    BITWORD * const our_after = after[word];

    /*  Do: after[word] = before[word] + ourmask[word] */
    bitpar_copy(our_after, before[word], s_bitpar_levs);
    bitpar_add_onerow(ourmask[word],our_after);
    if (bitpar_find_max(our_after, s_bitpar_levs) > s_q)
      return(0);          /* failed */
  }
  return(1);              /* passed */
}

/*  Returns 0 if tuple failed the q constraint.  Returns 1 if passed.
    If we return 1 then set up qvecs_ret
*/
static int
neg_update(
BITWORD * const * const masks,        /* [ch][trans word] */
BITWORD ** const * const oldnegcount, /* [ch][trans word][bitpar word] */
BITWORD ** const * const newnegcount  /* [ch][trans word][bitpar word] */
)
{
  int ch;                         /* character */
  for (ch = 0; ch < s_nnegchars; ch++) {
    if (! neg_update_inner(s_neg_ntrans_words[ch],
                           masks[ch],oldnegcount[ch],newnegcount[ch]))
      return(0);
  }
  return(1);
}

/*  Update based on adding a pixel at level "lev" at "piidc.
    do_tupdate() must have been called for the previous parts of
    the partial solution.  (We use static variables, and this
    function has side effects.)

    Returns 0 if failure, 1 if success.
*/
int
do_tupdate(
const int lev,
const int piidc
)
{
  const int piid  = PIIDC_TO_PIID(piidc);
  const int color = PIIDC_TO_COLOR(piidc);

  if (! pos_update(s_pos_tmask->mask[piid][color], s_pos_mask[lev-1],
                                  s_pos_mask[lev]))
                  /* also does positive evaluation, since it is so cheap */
    return(0);

  if (! neg_update( s_neg_tmask->mask[piid][color],
                    s_neg_count[lev-1], s_neg_count[lev]))
    return(0);

  /* set things used by do_teval */
  ASSERT(piidc >= 0);     /* conversion not well defined if piidc<0 */
  s_ps[lev]     = piidc;  /* conversion from signed to unsigned */
  s_lastlev     = lev;
  return(1);
}

/*  Sets qvec_ret[] according to the evaluation of after.
*/
static void
neg_evaluate_inner(
const int maxmatch,             /* MIN(s_q, partial solution length) */
const int nwords,               /* num. of translation words */
BITWORD * const * const after,  /* [tword][bitpar word] */
int * const qvec_ret            /* [0 .. QVEC_ROWS-1] */
)
{
  int word;

  ASSERT(maxmatch-(QVEC_ROWS-1) >= 0);
  vec_zero(qvec_ret, QVEC_ROWS);
  for (word = 0; word < nwords; word++) {
    const BITWORD * const our_after = after[word];
    int i;
    for (i = 0; i < QVEC_ROWS; i++) {
      const unsigned v = maxmatch-i;
      /* count how many columns in "our_after" have value v */
      const int count = bitpar_count_columns(our_after,s_bitpar_levs,v);
      qvec_ret[i] += count;
    }
  }
}

static void
neg_evaluate(
const int lev,
BITWORD ** const * const newnegcount, /* [ch][trans word][bitpar word] */
int * const * const qvecs_ret/* [qval: 0 .. QVEC_ROWS-1][0 .. s_nnegchars-1] */
                             /* rows sorted in ascending order */
)
{
  /*  COULDDO: QVEC_ROWS is a constant; I think bsbegin/bsend can be
      eliminated here in favor if a fixed size array.  AS 13Sep95.
  */
  bsbegin();
  {
    int ch;
    int * const qbuff = vec_alloc(QVEC_ROWS, bsalloc);
    const int maxmatch = MIN(s_q, lev+1);

    /*  Set up qvecs.
    */
    for (ch = 0; ch < s_nnegchars; ch++) {
      int i;
      neg_evaluate_inner(maxmatch,s_neg_ntrans_words[ch],newnegcount[ch],qbuff);
      for (i = 0; i < QVEC_ROWS; i++)
        qvecs_ret[i][ch] = qbuff[i];
    }

    /*  Sort each row of qvecs_ret into descending order.
    */
    if (s_nnegchars > 1) {
      int row;
      for (row = 0; row < QVEC_ROWS; row++) {
        isort(qvecs_ret[row], s_nnegchars, sizeof(qvecs_ret[0]),
          int_compare_descending);
      }
    }
  }
  bsend();
}

void
do_teval(
struct teval_struct * const teval_ret
)
{
  /*  p values from latest evaluation are in s_pvals.
      Sort p values into ascending order.
      Use insertion sort since s_nposchars is probably small, i.e., < 10.
      COULDDO: use intsort.  faster, more complicated and error prone.
  */
  if (s_nposchars > 1)
    isort(s_pvals, s_nposchars, sizeof(s_pvals[0]), int_compare);
  vec_copy(teval_ret->pvals, s_pvals, s_nposchars);

  neg_evaluate(s_lastlev, s_neg_count[s_lastlev], teval_ret->qvecs);

  teval_ret->tiebreak = vec_u_hash(s_ps, s_lastlev+1); /* TODO - test this */
  teval_ret->pslen    = s_lastlev+1;
  vec_u_copy(teval_ret->ps, s_ps, teval_ret->pslen);
}


/*  Evaluate a given partial solution, and set up the positive and
    negative masks up to level "ps_len-1".  Returns 0 if failure,
    1 if success.
    
    If successful, and teval_ret is not NULL, we return the evaluation of
    the partial solution, in teval_ret.
*/
int
teval_from_scratch(
const int * const ps,                 /* partial solution in terms of piidc's */
const int ps_len,                     /* length of partial solution */
struct teval_struct * const teval_ret /* evaluation of p.s. (we set) */
                                      /* if NULL we do not set */
)
{
  int lev;

  ASSERT(ps_len > 0);

  for (lev = 0; lev < ps_len; lev++) {
    const int piidc   = ps[lev];
    const int passed  = do_tupdate(lev,piidc);
    if (! passed)
      return(0);                /* failed */
  }
  if (teval_ret)
    do_teval(teval_ret);
  return(1);                    /* passed */
}


/*
  ratio_compare(): return the sign of

              n1                           n2
      ------------------     -     ------------------   ,
          d1+epsilon                   d2+epsilon
  
  where epsilon is considered to be infinitesimally small.
  Here n1,d1,n2,d2 are assumed to be nonnegative.
*/

/* macros for constructing OUR_MAX based on DBL_DIG */
#define OURV(xx)    ( 1e ## xx )
#define OURW(xx)    OURV(xx)
#define OUR_MAX     OURW(DBL_DIG)

static int
ratio_compare(
const int n1,
const int d1,
const int n2,
const int d2
)
{
  ASSERT(n1 >= 0);
  ASSERT(d1 >= 0);
  ASSERT(n2 >= 0);
  ASSERT(d2 >= 0);

  if ((n1==0) && (n2==0))
    return(0);
  if ((n1==0) && (n2!=0))
    return(-1);
  if ((n2==0) && (n1!=0))
    return(1);

  if ((d1==0) && (d2==0)) {
    if (n1>n2)  return(1);
    if (n2>n1)  return(-1);
    return(0);
  }

  if ((d1==0) && (d2!=0))
    return(1);
  if ((d2==0) && (d1!=0))
    return(-1);
  
  ASSERT(n1 <= (OUR_MAX/d2)-1);   /* n1*d2 is small enough */
  ASSERT(n2 <= (OUR_MAX/d1)-1);   /* n2*d1 is small enough */
  if (n1*((double)d2) > n2*((double)d1))
    return(1);
  if (n1*((double)d2) < n2*((double)d1))
    return(-1);

  return(0);
}


/*  qsort-style eval comparison function.

    Returns:
      -1    evaluation e1 is considered better than evaluation e2
      0     the partial solutions are the same n-tuple
      1     e2 is better than e1
    
    The result of a comparison does not depend in any way on p.
    back2 depends on this.

    COULDDO 5Oct95: allow the comparison of partial solutions that are not
    the same length.  This functionality is not needed right now because
    partial solutions that are compared are always the same length.
*/
int
teval_compare(
const struct teval_struct * const e1,
const struct teval_struct * const e2
)
{
  int i;
#ifdef DEBUG
  debug_teval("e1", e1);
  debug_teval("e2", e2);
#endif/*DEBUG*/

  for (i = 0; i < QVEC_ROWS; i++) {
    const int cmp = ratio_compare(e2->pvals[0],e2->qvecs[i][0],
            e1->pvals[0],e1->qvecs[i][0]);
D(printf("cmp=%d\n", cmp);)
    if (cmp != 0)
      return(cmp);
  }

  /* COULDDO - use number of candidates as tiebreak. */
  /* TODO - profile; then maybe only compute tiebreak fields when needed. */

  /* tiebreak field. */
  if (e1->tiebreak < e2->tiebreak)  return(-1);
  if (e1->tiebreak > e2->tiebreak)  return(1);

  /* Use vector comparison of partial solutions as a last resort.
  */
  ASSERT(e1->pslen == e2->pslen);
  return(vec_u_cmp(e1->ps, e2->ps, e1->pslen));
}

#ifdef OLD
int
teval_compare(
const struct teval_struct * const e1,
const struct teval_struct * const e2
)
{
  const int remain = s_n - e1->pslen;
  ASSERT(e2->pslen == e1->pslen);

  /*  qvecs.
  */
  {
    int row;
    for (row = 0; row < QVEC_ROWS; row++) {
      const int qval = s_q - row;   /* q-like value */
      if (qval + remain > s_q) {
        const int cmp = vec_cmp(e1->qvecs[row],e2->qvecs[row],s_nnegchars);
        if (cmp)  return(cmp);
      }
    }
  }

  /* pvals.
    If e1->pvals vector is lexicogaphically smaller than e2->pvals by virtue
    of position i, then the ith smallest pval of e1 is smaller (worse) than
    the ith smallest pval of e2.  So e1 is worse.
  */
  {
    const int compare = 0 - vec_cmp(e1->pvals, e2->pvals, s_nposchars);
    if (compare) return(compare);
  }

  /* tiebreak. */
  if (e1->tiebreak < e2->tiebreak)  return(-1);
  if (e1->tiebreak > e2->tiebreak)  return(1);

  return(0);
}
#endif /*OLD*/

#ifdef OLD2
/*  qsort-style eval comparison function.
    Returns:
      -1    evaluation e1 is considered better than evaluation e2
      0     the partial solutions are the same n-tuple
      1     e2 is better than e1
    
    The result of a comparison does not depend in any way on p.
    back2 depends on this.
*/
int
teval_compare(
const struct teval_struct * const e1,
const struct teval_struct * const e2
)
{
  int i;

#ifdef DEBUG
  debug_teval("e1", e1);
  debug_teval("e2", e2);
#endif/*DEBUG*/

  for (i = 0; i < QVEC_ROWS; i++) {
    const int p1        = e1->pvals[0];
    const int q1        = e1->qvecs[0][i];
    const int p2        = e2->pvals[0];
    const int q2        = e2->qvecs[0][i];
    /*
      We comparing p1/q1 to p2/q2.  Here p1,q1,p2,q2 are all nonnegative.
      If q1 or q2 is 0 then these ratios are undefined.  So what we do is
      compare p1/(q1+epsilon) to p2/(q2+epsilon).  Here epsilon=1/M.

      Suppose
        p1 / q1              >    p2 / q2 .
      Then if 1/M is small,
        p1 / (q1 + 1/M)      >    p2 / (q2 + 1/M)
        M p1 / (M q1 + 1)    >    M p2 / (M q2 + 1)
        M p1 (M q2 + 1)      >    M p2 ( M q1 + 1 )
        p1 (M q2 + 1)        >    p2 (M q1 + 1)
      
      A similar derivation holds for <, and also in the reverse direction.
    */
    const int M         = 256;
    ASSERT(p1*(((double)M)*q2+1) < INT_MAX);
    ASSERT(p2*(((double)M)*q1+1) < INT_MAX);
    {
      const long result = p1 * (M*q2 + 1)  -  p2 * (M*q1 + 1);
D(printf("cmp=%d\n", (result>0)?-1:(result<0));)
      if (result > 0)           /* big ratio is good */
        return(-1);
      else if (result < 0)
        return(1);
    }
  }

  /* COULDDO - use number of candidates as tiebreak */

  /* tiebreak field. */
  if (e1->tiebreak < e2->tiebreak)  return(-1);
  if (e1->tiebreak > e2->tiebreak)  return(1);

  /* Use vector comparison of partial solutions as a last resort.
     Do not ever return 0.
  */
  {
    const int cmp = vec_u_cmp(e1->ps, e2->ps, e1->pslen);
    ASSERT(e1->pslen == e2->pslen);
    ASSERT(cmp);
    return(cmp);
  }
}
#endif  /*OLD2*/

void
teval_setup_pq(
const int p,
const int q
)
{
  s_p = p;
  s_q = q;
}


/* Set up static variables.  We also do some memory allocation and
   initialization.  COULDDO: move some of the work out of this function
   and into another function that is called less frequently.  This
   probably isn't worth doing, though.
*/
void
teval_setup_invariants(
const int n,
const struct tmask_struct * const pos_tmask,
const struct tmask_struct * const neg_tmask,
void  *(*af) P((const int))
)
{
  /*  Simple variables
  */
  s_n                 = n;
  s_bitpar_levs       = bitpar_levels_needed(s_n);
  s_pos_tmask         = pos_tmask;
  s_neg_tmask         = neg_tmask;
  s_nposchars         = pos_tmask->nchars;
  s_nnegchars         = neg_tmask->nchars;
  s_pos_ntrans_words  = pos_tmask->ntrans_words;
  s_neg_ntrans_words  = neg_tmask->ntrans_words;
  s_pvals             = vec_alloc(s_nposchars, af);
  s_ps                = vec_u_alloc(s_n, af);

  /*  Positive masks
  */
  {
    int lev;  /* level */
    int ch;   /* character */

    s_pos_mask = (*af) (sizeof(s_pos_mask[0]) * (s_n+1));
    s_pos_mask++;       /* indexed as -1 .. s_n-1 */
    for (lev = -1; lev < s_n; lev++) {
      s_pos_mask[lev] = (*af) (sizeof(s_pos_mask[0][0]) * s_nposchars);
      for (ch = 0; ch < s_nposchars; ch++) {
        const int nbits = s_pos_tmask->ntrans[ch];
        s_pos_mask[lev][ch] = bit_alloc(nbits, af);
        if (lev == -1)
          bit_ones(s_pos_mask[lev][ch], nbits);
      }
    }
  }

  /*  Negative counts
      COULDDO: copy invariants out of loops
  */
  {
    int lev;    /* level */
    int ch;     /* character */
    int word;   /* translation word */

    s_neg_count = (*af) (sizeof(s_neg_count[0]) * (s_n+1));
    s_neg_count++;    /* indexed as -1 .. s_n-1 */
    for (lev = -1; lev < s_n; lev++) {
      s_neg_count[lev] = (*af) (s_nnegchars * sizeof(s_neg_count[0][0]));
      for (ch = 0; ch < s_nnegchars; ch++) {
        const int nwords = s_neg_ntrans_words[ch];
        s_neg_count[lev][ch] = (*af) (nwords * sizeof(s_neg_count[0][0][0]));
        for (word = 0; word < nwords; word++) {
          s_neg_count[lev][ch][word] = bitpar_alloc(s_bitpar_levs,af);
          if (lev == -1)
            bitpar_zero(s_neg_count[lev][ch][word], s_bitpar_levs);
        } /* word */
      }   /* ch */
    }     /* lev */
  }

  /* Note: candidates are set up in do_tupdate
  */
}
