/*
  -------------------- file tuputil.c
  Created 19 Nov 93 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Miscellaneous tuple functions that can be called by applications.
  Intended mostly for my own use, and not GN/Jung's.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "our_common.h"
#include "../libinclude/ntup.h"
#include "../libinclude/tuputil.h"


/*  Allocate space in the fields of a tuple structure, making the structure
    ready to hold tuples of size n.  Fills in "n" field of tuple.
*/
void
tuple_alloc_fields(
struct tuple_struct *tup,
const int n,
void  *(*allocfunc) P((const int))
)
{
  tup->pixels = (*allocfunc) (n * sizeof(tup->pixels[0]));
  tup->n = n;
}

/*  Deallocate space in the fields of a tuple structure.
*/
void
tuple_dealloc_fields(
struct tuple_struct *tup,
void  (*freefunc) P((void *))
)
{
  (freefunc) (tup->pixels);
}

void
pix_copy(
struct pix_struct *dest,
const struct pix_struct *src
)
{
  dest->row   = src->row;
  dest->col   = src->col;
  dest->color = src->color;
}

/*  Copy a tuple of size src->n.  Assumes destination has already been
    allocated to hold a tuple of size src->n.
*/
void
tuple_copy(
struct tuple_struct *dest,
const struct tuple_struct *src
)
{
  const int n = src->n;
  int i;

  ASSERT(dest->n == n);
  dest->p = src->p;
  dest->q = src->q;
  for (i = 0; i < n; i++) 
    pix_copy(& dest->pixels[i], & src->pixels[i]);
}

/*  Make an array of "len" tuples.  We set up each array element is a tuple
    structure that is is set up to hold data for an n-tuple.

    Return NULL if list length is 0.
*/
struct tuple_struct *
tuplelist_make(
const int listlen,
const int n,
void  *(*allocfunc) P((const int))
)
{
  struct tuple_struct *list;
  int i;

  if (listlen == 0)  return(NULL);

  list = (allocfunc) ( listlen * sizeof(list[0]) );
  for (i = 0; i < listlen; i++)
    tuple_alloc_fields(& list[i], n, allocfunc);
  return(list);
}

/*  Free the space associated with an array of "len" tuples.
    Does nothing if list is length 0.
*/
void
tuplelist_free(
struct tuple_struct *list,
const int listlen,
void  (*freefunc) P((void *))
)
{
  int i;

  if (listlen > 0) {
    for (i = 0; i < listlen; i++)
      tuple_dealloc_fields( & list[i], freefunc);
    (freefunc) (list);
  }
}
