/*
  -------------------- file memalloc.c

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef _020992_124244_memalloc
#define _020992_124244_memalloc

#include <malloc.h>
#include "common.h"

/* #define  BSDEBUG */
/* NOTE - changing BSDEBUG changes some constants; see memalloc.c.
   The constants are changed for the memory allocation test suite.
   You may want to change the constants manually if you are debugging
   an actual program and want to have BSDEBUG turned on and get
   the same behavior as when BSDEBUG is not turned on.
*/
/*#define PRINT_MEMORY_DEBUG*/
/* #define  MEMORY_DEBUG */

/*  Make sure that "xxnewlen" bytes are allocated at "xxp".  "xxoldlen"
    bytes were previously allocated.  Use this like a function that
    returns void.
*/

#define xalloc(xxp,xxnewlen,xxoldlen)       \
  do {                                      \
    if ((xxoldlen) < 1) {                   \
      if ((xxnewlen) > 0) {                 \
        (xxp) = xmalloc(xxnewlen);          \
        (xxoldlen) = (xxnewlen);            \
      }                                     \
      else {                                \
        (xxp) = xmalloc(1);                 \
        (xxoldlen) = 1;                     \
      }                                     \
    }                                       \
    else if ((xxnewlen) <= (xxoldlen)) ;    \
    else {                                  \
      (xxp) = xrealloc((xxp),(xxnewlen));   \
      (xxoldlen) = (xxnewlen);              \
    }                                       \
  } while (0)


/*  The "pool": array of buffers and their lengths */
struct _memalloc_pool_struct {
  unsigned char   *buffer;
  int             allocated;      /* how much allocated in buffer */
  int             used;           /* how much used in buffer */
};

#ifdef  BSDEBUG
#define BSD(xx)   do { xx } while (0) ;
void bsdebug P ((const char *str ));

#else
#define BSD(xx)   
#endif  /*BSDEBUG*/

/*  ALIGN_SIZE is assumed to be a power of 2 - if not then you need to
    change this to be a power of 2, as is appropriate for your machine
*/
#define ALIGN_SIZE  (sizeof(double))

/*  If we are going to allocate xx bytes, then align the request so that
    the request is for a number of bytes that satisfies the largest object
    needing alignment.

    Note: MEM_ALIGN_REQUEST may be used by user functions.  The user
          functions should not depend on the specific behavior of
          MEM_ALIGN_REQUEST, but user functions are allowed to depend
          on MEM_ALIGN_REQUEST performing its documented function as above.
    Note: users are allowed to depend on the following:
      MEM_ALIGN_REQUEST(x)+MEM_ALIGN_REQUEST(y) >= MEM_ALIGN_REQUEST(x+y)
    Note: ~(ALIGN_SIZE-1) masks the align bit and the bits to its left.
*/
#define MEM_ALIGN_REQUEST(xx) ( ((xx) + ALIGN_SIZE-1) & (~(ALIGN_SIZE-1)) )


/**************************************************************************
  inline functions (be careful about changing the order of func. definition)
**************************************************************************/

/* declarations (included for compatibility with non-gnuc compilers) */
#ifndef __GNUC__
void *xmalloc P((const int));
void *xcalloc P((const int));
void *xrealloc P((void *,const int));
void xfree P(( void * ));
extern void bsbegin();
extern void bsend();
extern void *bsalloc P((const int needed));
#endif  /*__GNUC__*/

/*  note: inline is defined to the empty string unless we're using gnuc */
#undef  INLINE
#ifdef  SRCFILE_LIBPROG_MEMALLOC
#define INLINE  __inline__
#else
#define INLINE  __inline__ extern
#endif

#if defined(__GNUC__) || defined(SRCFILE_LIBPROG_MEMALLOC)

/*  Note:  to free memory, use xfree() (defined in memalloc.h), and not
    free().  This aids debugging.
*/

int   printf();

#ifdef  PRINT_MEMORY_DEBUG
#define PMD(xx)   do { xx } while (0) ;
extern int    printf();
#else
#define PMD(xx)
#endif  /* PRINT_MEMORY_DEBUG */

INLINE void *
xmalloc(const int len)
{
  void  *p;
  p = (void *) malloc(len);

  PMD( (void) printf("xmalloc(%d) == %d\n", len, (int) p);)
  if (p) return p;
  fatal("xmalloc");
  return(NULL);
}

INLINE void *
xcalloc(const int len)
{
  void  *p;

  p = (void *) calloc(1, len);
  PMD( (void) printf("xcalloc(%d) == %d\n", len, (int) p);)
  if (p) return p;
  fatal("xmcalloc");
  return(NULL);
}

INLINE void *
xrealloc(
void *oldp,
const int len
)
{
  void  *p;

  p = (void *) realloc((char *) oldp, len);
  PMD( (void) printf("xrealloc(%d, %d) == %d\n", (int) oldp, len, (int) p);)
  if (p) return p;
  fatal("xrealloc");
  return(NULL);
}

#ifdef  MEMORY_DEBUG
INLINE void
xfree(
void  *oldp
)
{
  PMD ( (void) printf("xfree(%d)\n",  (int) oldp); )
  xrealloc(oldp,4);       /* test to see if oldp was really allocated */
  free(oldp);
}
#else   /*MEMORY_DEBUG*/
INLINE void
xfree(
void  *oldp
)
{
  PMD ( (void) printf("xfree(%d)\n",  (int) oldp); )
  free(oldp);
}
#endif  /*MEMORY_DEBUG*/

/*  The user calls this function to tell the allocation system that we are
    entering a region that makes use of the buffer stack.
*/
INLINE void
bsbegin()
{
  extern void   _memalloc_do_used_alloc P(());
  extern int    _memalloc_depth;
  extern int    _memalloc_used_alloc;
  extern int    *_memalloc_used;
  extern int    _memalloc_used_at_deepest;

  BSD(bsdebug("bsbegin(): top");)
  if (_memalloc_depth >= 0)
    _memalloc_used[_memalloc_depth] = _memalloc_used_at_deepest;
  _memalloc_depth++;
  if (_memalloc_depth >= _memalloc_used_alloc)
    _memalloc_do_used_alloc();
  _memalloc_used_at_deepest = 0;
  BSD(bsdebug("bsbegin(): bottom");)
}

/*  The user calls this function to request a space allocation.  This
    function is called between a bsbegin() and bsend() call.
*/
INLINE void *
bsalloc(
const int   needed
)
{
  const int get = MEM_ALIGN_REQUEST(needed);
  unsigned char *ret;

  extern void           _memalloc_nextpool P(( const int ));
  extern int            _memalloc_used_at_deepest;
  extern int            _memalloc_pool_used_at_deepest;
  extern int            _memalloc_pool_allocated_at_deepest;
  extern unsigned char  *_memalloc_pool_buffer_at_deepest;

  if (get == 0) {
    extern unsigned char  _memalloc_zerolenchar;

    BSD(printf("-------------------- bsalloc(%d aligned to %d) returns @%d\n",
      needed, get, (int) & _memalloc_zerolenchar);)
    return(& _memalloc_zerolenchar);
  }

  BSD(printf("---------- bsalloc(%d aligned to %d)\n", needed, get);)
  BSD(bsdebug("top of bsalloc()");)
  /* handles case when pool depth is -1 */
  if (_memalloc_pool_used_at_deepest+get > _memalloc_pool_allocated_at_deepest)
    _memalloc_nextpool(get);                      /* increments pool_depth */

  ASSERT(_memalloc_pool_used_at_deepest + get <=
                                        _memalloc_pool_allocated_at_deepest);
  _memalloc_used_at_deepest += get;
  ret = & _memalloc_pool_buffer_at_deepest[_memalloc_pool_used_at_deepest];
  _memalloc_pool_used_at_deepest += get;
#ifdef  BSDEBUG
  printf("-------------------- bsalloc(%d aligned to %d) returns @%d\n",
    needed, get, (int) ret);
  bsdebug("after bsalloc()");
#endif  /* BSDEBUG */
  return(ret);
}

/*  The user calls this function to tell the allocation system that we are
    leaving a region that makes use of the buffer stack.
*/
INLINE void
bsend()
{
  extern void   _memalloc_reclaim();
  extern int    *_memalloc_used;
  extern int    _memalloc_used_at_deepest;
  extern int    _memalloc_depth;
  extern int    _memalloc_pool_used_at_deepest;

  BSD(bsdebug("at beginning of bsend()");)
  if (_memalloc_used_at_deepest) {
    if (_memalloc_used_at_deepest >= _memalloc_pool_used_at_deepest)
      _memalloc_reclaim();
    else
      _memalloc_pool_used_at_deepest -= _memalloc_used_at_deepest;
  }
  ASSERT(_memalloc_depth >= 0);
  _memalloc_depth--;
  _memalloc_used_at_deepest = _memalloc_used[_memalloc_depth];
  BSD(bsdebug("at end of bsend()");)
}

#endif  /* defined(__GNUC__) || defined(source file) */

#endif /* _020992_124244_memalloc */
