/*
  -------------------- file vector.c
  Created Spring 92 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Routines for vector manipulation (see also matvec.c). 

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <limits.h>

#define SRCFILE_LIBPROG_VECTOR

#include "our_common.h"
#include "bitarray.h"
#include "getnum.h"
#include "io.h"
#include "matrix.h"
#include "memalloc.h"
#include "sort.h"
#include "util.h"
#include "vector.h"

/* #define HASHDEBUG */


/*  There are often two versions of a routine, one for matrices of
    bytes and one for matrices of arrays of ints.  The convention is
    to place the two routines next to each other in the source,
    with the character routine first.
*/

/*  Compare distributions of vectors v1 and v2.  Assume that the
    values of the elements of v1 and v2 are in the range min .. max
    ("min" and "max" do not have to be tight.)
    Returns 0 if the distributions match, otherwise 1.

    If retdistinct != NULL, and if the distributions match, then *retdistinct
    is set to 1 if all the elements of v1 (and therefore v2) are distinct,
    and 0 otherwise.

    ************* NOTE *************:  If retdistinct != NULL
    and the vectors don't match, then *retdistinct is undefined.

    Note:  allocates memory that is never freed.
*/
int
vec_cmp_distrib(
const int   *v1,
const int   *v2,
const int   n,
const int   min,
const int   max,
int         *retdistinct    /* are count vectors all ones? */
)
{
  static int    buffsize = 0;
  static int    *count1,*count2;
  int           range;      /* max-min+1 */
  int           i;
  int           distinct;   /* are all elements of v1 distinct? */

  range = (max-min+1);
  /* memory allocation */
  if (range > buffsize) {
    if (buffsize == 0) {
      count1      = xmalloc(sizeof(int) * range);
      count2      = xmalloc(sizeof(int) * range);
    }
    else {
      count1      = xrealloc(count1, sizeof(int) * range);
      count2      = xrealloc(count2, sizeof(int) * range);
    }
    buffsize = range;
  }

  /* computation of distributions */
  for (i = 0; i < range; i++)
    count1[i] = count2[i] = 0;
  for (i = 0; i < n; i++) {
    ASSERT((v1[i] >= min) && (v1[i] <= max));
    ASSERT((v2[i] >= min) && (v2[i] <= max));
    count1[v1[i]-min]++;
    count2[v2[i]-min]++;
  }
#ifdef  VDISTRIB_DEBUG
  printf("vec_distrib()\n");
  printf("v1 = ");
  vec_print(stdout,v1,n);
  printf("v2 = ");
  vec_print(stdout,v2,n);
  printf("count1 = ");
  vec_print(stdout,count1,n);
  printf("count2 = ");
  vec_print(stdout,count2,n);
#endif  /* VDISTRIB_DEBUG */

  /* comparison */
  distinct = 1;
  for (i = 0; i < range; i++) {
    if (count1[i] != count2[i])
      return(1);
    if (count1[i] > 1)
      distinct = 0;
  }
  if (retdistinct)
    *retdistinct = distinct;
  return(0);
}

/*  Print a vector.  Uses left margin of 1 and puts newline afterward.  */
void
vec_print(
FILE        *f,
const int   *v,
const int   n
)
{
  (void) vec_print2(f,v,n,1,1);
  printf("\n");
}

/*  Print a vector, but add "offset" to each element.
    Uses left margin of 1 and puts newline afterward.
*/
void
vec_print_offset(
FILE        *f,
const int   *v,
const int   n,
const int   offset
)
{
  int   *scratch;

  bsbegin();
  scratch = vec_alloc(n, bsalloc);
  vec_copy(scratch, v, n);
  vec_add_scalar(scratch, n, offset);
  vec_print(f,scratch, n);
  bsend();
}

void
vec_u_print(
FILE                  *f,
const unsigned int   *v,
const int             n
)
{
  (void) vec_u_print2(f,v,n,1,1);
  printf("\n");
}

void
vec_uchar_print(
FILE                  *f,
const unsigned char   *v,
const int             n
)
{
  (void) vec_uchar_print2(f,v,n,1,1);
  printf("\n");
}


/*  More complicated version of vec_print that allows left margin specification
    and specification of the initial column.  Does not put new line after
    vector.
*/
int
vec_print2(
FILE      *f,
const int *v,
const int n,
const int incol,
const int leftmargin
)
{
  int           min, max, width;

  /* determine field width */
  vec_minmax(v,n,&min,&max);
  width = MAX(fieldlen(min), fieldlen(max));

  return(list_print( f, v, sizeof(int), n, 0, 1, incol, leftmargin,
            OUTPUT_COLUMNS, width, &width, list_int_print, list_int_compare));
}

int
vec_u_print2(
FILE               *f,
const unsigned int *v,
const int           n,
const int           incol,
const int           leftmargin
)
{
  unsigned int  min,max;
  int           width;

  vec_u_minmax(v,n,&min,&max);
  width = MAX(fieldlen(min), fieldlen(max));

  return(list_print( f, v, sizeof(unsigned int), n, 0, 1, incol, leftmargin,
        OUTPUT_COLUMNS, width, &width, list_u_print, list_u_compare));
}

int
vec_uchar_print2(
FILE                *f,
const unsigned char *v,
const int           n,
const int           incol,
const int           leftmargin
)
{
  unsigned char min,max;
  int           width;

  vec_uchar_minmax(v,n,&min,&max);
  width = MAX(fieldlen(min), fieldlen(max));

  return(list_print( f, v, sizeof(unsigned char), n, 0, 1, incol, leftmargin,
        OUTPUT_COLUMNS, width, &width, list_uchar_print, list_uchar_compare));
}

void
vec_print_as_matrix(
FILE *f,
const int *v,
const int n,        /* matrix width (vector must be n*n or more) */
const unsigned char *to_row,
const unsigned char *to_col
)
{
  int **m;
  int i;

  m = mat_alloc(n, xmalloc);

  for (i = 0; i < n*n; i++)
    m [to_row[i]] [to_col[i]] = v[i];
  mat_print(f, m, n);
  mat_free(m, xfree);
}

void
vec_u_print_as_matrix(
FILE *f,
const unsigned char *v,
const int n,        /* matrix width (vector must be n*n or more) */
const int *to_row,
const int *to_col
)
{
  unsigned char **m;
  int i;

  m = mat_uchar_alloc(n, xmalloc);

  for (i = 0; i < n*n; i++)
    m [to_row[i]] [to_col[i]] = v[i];
  mat_uchar_print(f, m, n);
  mat_uchar_free(m, xfree);
}


int
vec_u_read(
FILE *f,
unsigned char   *v,
const int       n
)
{
  int   i,val;

  for (i = 0; i < n; i++) {
    if (! getnum(f, & val))
      return(0);
    v[i] = val;
  }
  return(1);
}

int
vec_read(
FILE            *f,
int             *v,
const int       n
)
{
  int   i,val;

  for (i = 0; i < n; i++) {
    if (! getnum(f, & val))
      return(0);
    v[i] = val;
  }
  return(1);
}

/*  Convert a vector to table form.  See also vec_table_to_vec.
    The vector must have already been sorted.
    The vector should be of size >=  max(vec) - min + 2
*/
void
vec_to_table(
const int   *vec,             /* vector to convert into table format */
const int   n,                /* vector length */
const int   min,              /* first entry in table corresponds to min */
const int   end_of_row_mark,  /* end of table is marked with this value */
int         *table_ret        /* table is returned here */
)
{
  int       i,t,val,j;

  val = min;
  i = j = t = 0;
  while (i < n) {
    ASSERT((i>=n-1) || (vec[i] <= vec[i+1]));   /* vector must be presorted */
    if (vec[i] != val) {
      table_ret[t++] = j;
      val++;
      j = i;
    }
    else
      i++;
  }
  table_ret[t++] = j;
  table_ret[t++] = end_of_row_mark;       /* mark end of row */
}

/*  Convert from table format to vector format.  Example:
      min == -1
      end_of_row_mark = -2
      n == 12
      table == [ 1 3 4 6 6 8 end_of_row_mark X X ]
      table_size = 8 - (-1) + 2 = 11
      result:            vret == [ 0 0 0 1 1 2 3 3 5 5 -2 ]
       (corresponding to values   -1 0 1 2 3 4 5 6 7 8 )
    
    The returned vector has the property that it is sorted in ascending
    order.
*/
void
vec_table_to_vec(
const int   *table,
const int   min,
const int   n,                /* vector length */
const int   end_of_row_mark,
int         *vret
)
{
  int   i,t;        /* index into vector and table */
  int   c;          /* how many entries to fill in for this table entry */
  int   val;        /* value corresponding to this table entry */

  for (i = 0, t = 0, val = min; i < n; t++, val++) {
    if (table[t+1] == end_of_row_mark)
      c = n - i;
    else
      c = table[t+1] - table[t];
    while (c--) {
      ASSERT(i<n);
      vret[i++] = val;
    }
  }
}


/*
  This function implements a 32-bit linear feedback shift register with
	taps at bits 32,7,6,2 (offset 1).  Doing w = lfsr32(w) is equivalent to
    "for(i=0;i<32;i++)w=NEXT(w);",
	where w is a word whose rightmost 32 bits are arbitrary and all
	other bits of w (if any) are 0.  (NEXT() is defined below.)  This
  function uses about 81 SPARC instructions, as compared to about 330
  for the NEXT approach.  This function was tested for all 2^32 32-bit
  quantities versus the brute force approach, and both approaches gave
	the same answer for all values.

  Should be ANSI-portable.
	(Note: NEXT() may not be ANSI portable because of its dependence
	on a 32-bit word size).
*/

#ifdef UNUSED
/* Generate next element in the LFSR using taps of (32,7,6,2) (1-offset).  */
#define NEXT(y)  (((y)<<1) | (1 & (((y)>>31)^((y)>>6)^((y)>>5)^((y)>>1))))
#endif /*UNUSED*/

static unsigned
lfsr32(
const unsigned w
)
{
  unsigned b;
  unsigned r,s;

  /* 32,31 */   b = (w ^ (w << 25) ^ (w << 26) ^ (w << 30)) & 0xffffffff;
                r = (b & 0xc0000000);
  /* 30,29 */   b ^= r >> 2;
  /* 28,27 */   b ^= (b >> 2) & 0x0c000000;
                s = b & 0x0c000000;

  /* 26,25 */   b ^= (r ^ (s << 4)) >> 6;
  /* 26 completed */
  /* 25-19 */   b ^= (b >> 7) & 0x01fc0000;
  /* 25 completed */
  /* 24-19 */   b ^= (b >> 6) & 0x00fc0000;
  /* 24,23 */   b ^= (b >> 2) & 0x00c00000;
  /* 22,21 */   b ^= (b & 0x00c00000) >> 2;
  /* 20,19 */   b ^= (b >> 2) & 0x000c0000;

                b ^= ((b ^ (b >> 1)) & 0x00fc0000) >> 6;
  /* 18,17 */   b ^= (b & 0x000c0000) >> 2;
  /* 16,15 */   b ^= (b >> 2) & 0x0000c000;
  /* 14,13 */   b ^= (b & 0x0000c000) >> 2;

                b ^= ((b >> 6) ^ (b >> 7)) & 0x00000fc0;
  /* 12,11 */   b ^= (b >> 2) & 0x00000c00;
  /* 10,9  */   b ^= (b & 0x00000c00) >> 2;
  /* 8,7   */   b ^= (b >> 2) & 0x000000c0;

                b ^= ((b >> 6) ^ (b >> 7)) & 0x0000003f;
  /* 6,5   */   b ^= (b >> 2) & 0x00000030;
  /* 4,3   */   b ^= (b >> 2) & 0x0000000c;
  /* 2,1   */   b ^= (b >> 2) & 0x00000003;

  return(b);
}


/*  Given a vector of integers, return an unsigned integer (its "hash")
    that is pseudo-random and dependent on the vector only.  This is
    completely deterministic.  It is hoped that two vectors that are
    permutations of each other will not hash to the same value.

    MACHINE DEPENDENT.  Gives fatal error unless sizeof(unsigned)==4

    COULDDO: This provides a method of comparing unsorted vectors
    to see if their sorted versions are equal (set equality).
    (We would use a modified verison of vec_u_hash where the order
    of the vector elements was insignificant.)  If the hashes are
    unequal, then the vectors are unequal; otherwise, sort them and
    compare them.

    This function has the property that it maps [0,2^32] onto [1,2^32],
    i.e., there is only one collision.  The values RAND1 and (RAND1^RAND2)
    are the only values that collide.

    COULDDO: it may be possible to speed up the computation of several
    consecutive applications of NEXT, by reusing intermediate results.
*/


unsigned
vec_u_hash(
const unsigned * const vec,
const int n
)
{
  const unsigned RAND1=0x39721001;  /* random 32-bit string (from random()) */
  const unsigned RAND2=0x7acee6ed;  /* random 32-bit string (from random()) */
  int i;
  unsigned w;

#ifdef HASHDEBUG
  {
    int j;
    for (j = 0; j < n; j++) {
      printf(" %08x@%d\n", vec[j],j);
    }
  }
#endif  /*HASHDEBUG*/

  if (sizeof(w) != 4)
    fatal("word size not handled by vec_u_hash()");

  for (w = RAND1, i = 0; i < n; i++) {
    const unsigned v = vec[i];
    w = w ^ v;
    if (w == 0) {                       /* we hope that this happens rarely */
      if (v == RAND1)  w = RAND2;
      else             w = v ^ RAND1;
    }
    w = lfsr32(w);
    w = lfsr32(w);
  }

  return(w-1);      /* w cannot be 0; we return value from 0 .. UINT_MAX-1 */
}
#ifdef OLD

/*  Do (a+b) mod UINT_MAX in unsigned arithmetic without overflow.
*/
static unsigned
our_uadd(
unsigned a,
unsigned b
)
{
  if (b <= UINT_MAX-a)            /* a+b <= UINT_MAX */
    return(a+b);
  else
    return(UINT_MAX - ((UINT_MAX-a)+(UINT_MAX-b)));   /* a+b-UINT_MAX */
}
#define NHASHMASKS 59     /* may be good to make this a prime number */

/* cloned in randpack */
unsigned
vec_u_hash(
const unsigned * const vec,
const int n
)
{
  /* table of random bit strings */
  static BITWORD randmasks[NHASHMASKS] = {
    0x39721001, 0x7acee6ed, 0x4a5b2474, 0xa1d58d53, 0x7d695da4, 0x8e3ec2d9,
    0xca566521, 0x61ee643e, 0x910e6247, 0x9e6d2476, 0x6d5da7e1, 0x0d87a364,
    0xbd427011, 0xe808477b, 0xcc974015, 0xe1494257, 0x01171318, 0xb1870035,
    0xccc9d564, 0x62524a14, 0xc3356459, 0xce1d6534, 0x975e53a3, 0x57693e57,
    0x20405a99, 0x13e70939, 0x281e3b20, 0xb01606d2, 0x79a4bbdc, 0xe55ee818,
    0x927c5ad1, 0xa6ead45c, 0xe78a3d83, 0x3d2363b7, 0xc45aea31, 0x134671a6,
    0x839b92ec, 0x790544d2, 0x8a4ec775, 0x2b4ace85, 0xa994388a, 0x918e5d85,
    0x8cbcba2e, 0x761bd190, 0xb3be342d, 0xcac90560, 0x92c5debc, 0xdc03e932,
    0x4e18aad0, 0xca9c71c1, 0xa9601346, 0x7596e404, 0xa24c42c2, 0x4eeb04c2,
    0x5a3d574d, 0xc5b41c03, 0x641ad5c2, 0x44db82e5, 0xc243181e
  };
  /* random permutations of 0 .. 31 */
  static int p1[32] =
          {  5,  7,  0, 30, 17, 26, 23,  3,  9,  4, 25, 18, 29, 19, 21,  6,
            14, 10, 31, 24, 22, 11, 28,  2, 16, 20,  8,  1, 13, 15, 12, 27 };
  static int p2[32] =
          { 25, 29, 8, 18, 26,  1, 14,  6, 30, 28, 27,  2, 20, 16, 19,  7,
            22,  3, 0, 17, 31, 12, 10, 21,  4, 13,  5, 11,  9, 23, 15, 24 };

  BITWORD x,y;
  int i;

  if (BITS_PER_WORD != 32)
    fatal("word size not handled by vec_u_hash()");

  for (x = y = ALL_BITS_ZERO, i = 0; i < n; i++) {
    const unsigned v    = vec[i];
    unsigned v1         = our_uadd(v,564385021u);
    unsigned v2         = our_uadd(v,1831u);
    BITWORD w1,w2;    /* BITWORD is really same as unsigned */

    bit_permute(& w1, & v1, sizeof(BITWORD), p1);
    bit_permute(& w2, & v2, sizeof(BITWORD), p2);
    /*  w1 and w2 have now been computed solely from v.
        We hope they are scrambled versions of v.
    */
    our_uadd(x, w1);
    our_uadd(y, w2);
    x ^= randmasks[v1 % NHASHMASKS];
    y ^= randmasks[v2 % NHASHMASKS];
  }
#ifdef HASH_DEBUG
  {
    const BITWORD t = x^y;
    printf("vec_u_hash: ");
    bit_print_binary(stdout,& t, 32);
    printf(" from ");
    vec_u_print(stdout, vec, n);
  }
#endif  /*HASH_DEBUG*/

  return(x^y);
}
#endif /*OLD*/

/*  Version of vec_u_hash for unsigned characters.  Converts
    the string to unsigned longs, then calls vec_u_hash.
*/
unsigned
vec_uchar_hash(
const unsigned char * const uchar_vec,
const int n
)
{
  unsigned w;

  bsbegin();
  {
    unsigned int * const uvec = vec_u_alloc(n,bsalloc);
    int i;

    for (i = 0; i < n; i++)
      uvec[i] = uchar_vec[i];
    w = vec_u_hash(uvec, n);
  }
  bsend();
  return(w);
}

unsigned
vec_hash(
const int *const ivec,
const int n
)
{
  unsigned w;

  bsbegin();
  {
    unsigned int * const uvec = vec_u_alloc(n,bsalloc);
    int i;

    /* Convert int to unsigned.  Assumes that uvec[i] and vec[i] are the
       same size.  If so, the conversation is well defined and successful
       according to ANSI C.
    */
    for (i = 0; i < n; i++) {
      uvec[i] = ivec[i];
      /* TODO - test/scrutinize what happens when ivec has a negative value */
    }
    w = vec_u_hash(uvec, n);
  }
  bsend();
  return(w);
}
